@@ -35,6 +35,7 @@
 ### Valerio Bontempi <v.bontempi@inwind.it>: IPv6 support
 ### Lorenzo Colitti <lorenzo@colitti.com>: IPv6 support
 ### Philippe Simonet <Philippe.Simonet@swisscom.com>: Export avoid...
+### Luc Pauwels <Luc.Pauwels@xalasys.com>: use_16bit_request_ids
 ######################################################################
 
 package SNMP_Session;		
@@ -45,7 +46,8 @@ use strict;
 use Exporter;
 use vars qw(@ISA $VERSION @EXPORT $errmsg
 	    $suppress_warnings
-	    $default_avoid_negative_request_ids);
+	    $default_avoid_negative_request_ids
+	    $default_use_16bit_request_ids);
 use Socket;
 use BER '0.95';
 use Carp;
@@ -56,7 +58,7 @@ sub map_table_start_end ($$$$$$);
 sub index_compare ($$);
 sub oid_diff ($$);
 
-$VERSION = '0.99';
+$VERSION = '1.00';
 
 @ISA = qw(Exporter);
 
@@ -107,6 +109,13 @@ my $default_max_repetitions = 12;
 ###
 $SNMP_Session::default_avoid_negative_request_ids = 0;
 
+### Default value for "use_16bit_request_ids".
+###
+### Set this to non-zero if you have agents that use 16bit request IDs,
+### and don't forget to complain to your agent vendor.
+###
+$SNMP_Session::default_use_16bit_request_ids = 0;
+
 ### Whether all SNMP_Session objects should share a single UDP socket.
 ###
 $SNMP_Session::recycle_socket = 0;
@@ -125,7 +134,7 @@ BEGIN {
     $SNMP_Session::ipv6available = 0;
 
     if (eval {require Socket6;} &&
-	eval {require IO::Socket::INET6; IO::Socket::INET6->VERSION("2.00");}) {
+	eval {require IO::Socket::INET6; IO::Socket::INET6->VERSION("1.26");}) {
 	import Socket6;
 	$ipv6_addr_len = length(pack_sockaddr_in6(161, inet_pton(AF_INET6(), "::1")));
 	$SNMP_Session::ipv6available = 1;
@@ -183,6 +192,8 @@ sub encode_request_3 ($$$@) {
 	? -0x80000000 : $this->{request_id}+1;
     $this->{request_id} += 0x80000000
 	if ($this->{avoid_negative_request_ids} && $this->{request_id} < 0);
+    $this->{request_id} &= 0x0000ffff
+	if ($this->{use_16bit_request_ids});
     foreach $_ (@{$encoded_oids_or_pairs}) {
       if (ref ($_) eq 'ARRAY') {
 	$_ = &encode_sequence ($_->[0], $_->[1])
@@ -679,6 +690,7 @@ sub open {
 	   'lenient_source_address_matching' => 1,
 	   'lenient_source_port_matching' => 1,
 	   'avoid_negative_request_ids' => $SNMP_Session::default_avoid_negative_request_ids,
+	   'use_16bit_request_ids' => $SNMP_Session::default_use_16bit_request_ids,
 	   'capture_buffer' => undef,
 	  };
 }
